import bson
import os

from tempfile import mkdtemp
from shutil import rmtree

from django.test.utils import override_settings
from django.utils.http import urlencode
from mock import Mock

from core.wsgi import make_app, BzrRevnoMiddleware
from core.tests.factory import TestCaseWithFactory


@override_settings(ENABLE_ARTIFICIAL_OOPS=True)
class WSGIDispatchTestCase(TestCaseWithFactory):

    def setUp(self):
        """Create a temporary directory for oopses."""
        self.oops_dir = mkdtemp()
        self.addCleanup(rmtree, self.oops_dir)

        self.environ = {
            'PATH_INFO': '/',
            'QUERY_STRING': '',
            'REMOTE_ADDR': '127.0.0.1',
            'REQUEST_METHOD': 'GET',
            'SCRIPT_NAME': '',
            'SERVER_NAME': 'testserver',
            'SERVER_PORT': '80',
            'SERVER_PROTOCOL': 'HTTP/1.1',
            'wsgi.version': (1, 0),
            'wsgi.url_scheme': 'http',
            'wsgi.multiprocess': True,
            'wsgi.multithread': False,
            'wsgi.run_once': False,
            'wsgi.input': '',
            'oops.report': {},
        }
        super(WSGIDispatchTestCase, self).setUp()

    def request_500_exception(self, data={}):
        """Helper for creating a 500 exception."""
        self.environ['PATH_INFO'] = '/reviews/die/'
        if data != {}:
            self.environ['QUERY_STRING'] = urlencode(data, doseq=True)

        response = self.request_wsgi_response(self.environ)
        return response

    def request_wsgi_response(self, environ):
        start_response = Mock()
        oops_config = {
            'publishers': [{
                'inherit_id': True,
                'error_dir': self.oops_dir,
                'type': 'datedir',
                'new_only': False,
            }]}
        with override_settings(OOPSES=oops_config):
            app = make_app()
            response = app(environ, start_response).next()
        return response

    def read_oops_from_disk(self):
        # Helper to read in the oops written to disk.
        # There is only one date directory in the oops directory.
        dated_dirs = os.listdir(self.oops_dir)
        self.assertEqual(1, len(dated_dirs))

        # There is only one oops.
        todays_oops_path = dated_dirs[0]
        oops_list = os.listdir(self.oops_dir + '/' + todays_oops_path)
        self.assertEqual(1, len(oops_list))

        oops_path = self.oops_dir + '/' + todays_oops_path + '/' + oops_list[0]
        with open(oops_path, "r") as oops_file:
            oops_file_contents = oops_file.read()
        result = bson.decode_document(oops_file_contents, 0)
        if len(result) == 2 and isinstance(result[1], dict):
            return result[1]
        raise ValueError('oops file content is not correct')

    def test_oopsid_in_500_response(self):
        # The rendered response for a 500 exception includes the oops id.
        response = self.request_500_exception()

        self.assertTrue(self.environ['oops.report']['id'] in response)

    def test_traceback_in_500_oops_on_disk(self):
        # A request resulting in a 500 writes the oops to disk.
        self.request_500_exception()

        oops = self.read_oops_from_disk()
        self.assertEqual(oops.get('type'), 'ZeroDivisionError')
        self.assertIn('tb_text', oops)
        self.assertTrue(oops['tb_text'].startswith('  File '))

    def test_delay_triggers_an_oops(self):
        self.environ['PATH_INFO'] = '/reviews/delay/0.11/'
        with override_settings(SOFT_TIMEOUT_MILLISECONDS=1):
            self.request_wsgi_response(self.environ)
        oops_content = self.read_oops_from_disk()
        self.assertEqual('SoftRequestTimeout', oops_content['type'])
        expected = u'Start_response over timeout 1.'
        self.assertEqual(expected, oops_content['value'])

    def test_bzr_header(self):
        BzrRevnoMiddleware.REVNO_HEADER = ('HEADER', 'VALUE')
        app = make_app()
        _start_response = Mock('/',)
        app(self.environ, _start_response).next()
        start_response_args = _start_response.mock_calls[0][1]
        headers = start_response_args[1]
        self.assertIn(('HEADER', 'VALUE'), headers)
